#
#    Copyright 2017 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Model functionality
"""
from .client import ModelClient, load_image
from .types import int8, uint8, int16, uint16, int32, uint32, int64, uint64
from .types import float32, float64
from .types import string, boolean, image

import legion.pymodel.model
import legion.k8s.properties
from legion.utils import normalize_name

MODEL_TYPES = [
    legion.pymodel.model.Model
]


# Global variable for storing model context
# This variable initializes by init() function
_model = None

# Global variable for storing properties for current model
# This variable initializes by set_properties() function that calls on model init or before model deploy
properties = None  # type: legion.k8s.properties.K8SPropertyStorage


def get_context():
    """
    Get current model context

    :return: object -- model context
    """
    return _model


def reset_context():
    """
    Drop current model context

    :return: None
    """
    global _model
    _model = None


def set_properties(new_properties):
    """
    Set new properties as a instance of legion.k8s.properties.K8SPropertyStorage class

    :param new_properties: new properties
    :type new_properties: :py:class:`legion.k8s.properties.K8SPropertyStorage`
    :return: None
    """
    global properties
    properties = new_properties


def reset_properties():
    """
    Reset properties

    :return: None
    """
    global properties
    properties = None


def init(model_id, model_version, model_type=legion.pymodel.model.Model.NAME):
    """
    Initialize new model context

    :param model_id: model name
    :type model_id: str
    :param model_version: model version
    :type model_version: str
    :param model_type: type of model, one of MODEL_TYPES names
    :type model_type: str
    :return: object -- instance of Model class
    """
    global _model

    if _model:
        raise Exception('Context already has been defined')

    model_id = normalize_name(model_id)
    if not model_id:
        raise Exception('Model name string length should be greater that 1 (after normalization)')

    builder = [m_type for m_type in MODEL_TYPES if m_type.NAME == model_type]

    if not builder:
        raise Exception('Cannot find model builder for type {}'.format(model_type))

    if len(builder) > 1:
        raise Exception('More then 1 builder have been found for type {}'.format(model_type))

    _model = builder[0](model_id, model_version)
    set_properties(_model.properties)

    return _model


def send_metric(metric, value):
    """
    Send build metric value

    :param metric: metric type or metric name
    :type metric: :py:class:`legion.metrics.Metric` or str
    :param value: metric value
    :type value: float or int
    :return: None
    """
    if not _model:
        raise Exception('Context has not been defined')

    return _model.send_metric(metric, value)


def export_df(apply_func, input_data_frame, *, prepare_func=None, endpoint='default'):
    """
    Export simple Pandas DF based model as a bundle

    :param apply_func: an apply function DF->DF
    :type apply_func: func(x) -> y
    :param input_data_frame: pandas DF
    :type input_data_frame: :py:class:`pandas.DataFrame`
    :param prepare_func: a function to prepare input DF->DF
    :type prepare_func: func(x) -> y
    :param endpoint: (Optional) endpoint name, default is 'default'
    :type endpoint: str
    :return: model container
    """
    if not _model:
        raise Exception('Context has not been defined')

    return _model.export_df(apply_func, input_data_frame, prepare_func=prepare_func, endpoint=endpoint)


def export(apply_func, column_types, *, prepare_func=None, endpoint='default'):
    """
    Export simple parameters defined model as a bundle

    :param apply_func: an apply function DF->DF
    :type apply_func: func(x) -> y
    :param column_types: result of deduce_param_types or prepared column information
    :type column_types: dict[str, :py:class:`legion.model.types.ColumnInformation`]
    :param prepare_func: a function to prepare input DF->DF
    :type prepare_func: func(x) -> y
    :param endpoint: (Optional) endpoint name, default is 'default'
    :type endpoint: str
    :return: model container
    """
    if not _model:
        raise Exception('Context has not been defined')

    return _model.export(apply_func, column_types, prepare_func=prepare_func, endpoint=endpoint)


def export_untyped(apply_func, *, prepare_func=None, endpoint='default'):
    """
    Export simple untyped model as a bundle

    :param apply_func: an apply function DF->DF
    :type apply_func: func(x) -> y
    :param prepare_func: a function to prepare input DF->DF
    :type prepare_func: func(x) -> y
    :param endpoint: (Optional) endpoint name, default is 'default'
    :type endpoint: str
    :return: model container
    """
    if not _model:
        raise Exception('Context has not been defined')

    return _model.export_untyped(apply_func, prepare_func=prepare_func, endpoint=endpoint)


def save(path=None):
    """
    Save model to path (or deduce path)

    :param path: (Optional) target save name
    :return: model container
    """
    if not _model:
        raise Exception('Context has not been defined')

    return _model.save(path)


def define_property(name, initial_value):
    """
    Define model property

    :param name: property name
    :type name: str
    :param initial_value: initial property value
    :type initial_value: any
    :return: model container
    """
    if not _model:
        raise Exception('Context has not been defined')

    return _model.define_property(name, initial_value)



#
#    Copyright 2017 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Deploy logic for legion
"""

import logging
import os
import time
import re

import docker
import docker.errors
import legion.config
import legion.containers.docker
import legion.containers.headers
import legion.k8s
import legion.external.edi
import legion.external.grafana
import legion.pymodel
import legion.model
import legion.utils
from legion.utils import Colors, ExternalFileReader

LOGGER = logging.getLogger(__name__)

INSPECT_FORMAT_COLORIZED = 'colorized'
INSPECT_FORMAT_TABULAR = 'column'
VALID_INSPECT_FORMATS = INSPECT_FORMAT_COLORIZED, INSPECT_FORMAT_TABULAR


def build_model(args):
    """
    Build model

    :param args: command arguments
    :type args: :py:class:`argparse.Namespace`
    :return: :py:class:`docker.model.Image` docker image
    """
    client = legion.containers.docker.build_docker_client(args)

    with ExternalFileReader(args.model_file) as external_reader:
        if not os.path.exists(external_reader.path):
            raise Exception('Cannot find model binary {}'.format(external_reader.path))

        container = legion.pymodel.Model.load(external_reader.path)
        model_id = container.model_id

        image_labels = legion.containers.docker.generate_docker_labels_for_image(external_reader.path, model_id, args)

        LOGGER.info('Building docker image...')
        image = legion.containers.docker.build_docker_image(
            client,
            model_id,
            external_reader.path,
            image_labels,
            args.docker_image_tag
        )

        LOGGER.info('Image has been built: {}'.format(image))

        legion.utils.send_header_to_stderr(legion.containers.headers.IMAGE_TAG_LOCAL, image.id)

        if args.push_to_registry:
            legion.containers.docker.push_image_to_registry(client, image, args.push_to_registry)

        return image


def inspect_kubernetes(args):
    """
    Inspect kubernetes

    :param args: command arguments with .namespace
    :type args: :py:class:`argparse.Namespace`
    :return: None
    """
    edi_client = legion.external.edi.build_client(args)
    model_deployments = edi_client.inspect(args.model_id, args.model_version)

    data = []

    for deployment in model_deployments:
        if deployment.status == 'ok' and deployment.model_api_ok:
            line_color = Colors.OKGREEN
        elif deployment.status == 'warning':
            line_color = Colors.WARNING
        else:
            line_color = Colors.FAIL

        errors = ''

        if not deployment.model_api_ok:
            errors = 'MODEL API DOES NOT RESPOND'

        if errors:
            errors = 'ERROR: {}'.format(errors)

        data.append({
            'deployment': deployment,
            'errors': errors,
            'line_color': line_color
        })

    if args.format == INSPECT_FORMAT_COLORIZED:
        if data:
            print('%sModel deployments:%s' % (Colors.BOLD, Colors.ENDC))

            for item in data:
                arguments = (
                    item['line_color'], Colors.ENDC,
                    Colors.UNDERLINE, item['deployment'].model, Colors.ENDC,
                    item['deployment'].image, item['deployment'].version,
                    item['line_color'], item['deployment'].ready_replicas, item['deployment'].scale, item['errors'],
                    Colors.ENDC
                )
                print('%s*%s %s%s%s %s (version: %s) - %s%s / %d pods ready %s%s' % arguments)

        if not model_deployments:
            print('%s-- cannot find any model deployments --%s' % (Colors.WARNING, Colors.ENDC))

        if not data and model_deployments:
            print('%s-- cannot find any model deployments after filtering --%s' % (Colors.WARNING, Colors.ENDC))
    elif args.format == INSPECT_FORMAT_TABULAR:
        headers = 'Model ID', 'Image', 'Version', 'Ready', 'Scale', 'Errors'
        items = [[
            item['deployment'].model,
            item['deployment'].image,
            item['deployment'].version,
            str(item['deployment'].ready_replicas),
            str(item['deployment'].scale),
            item['errors'],
        ] for item in data]

        if data:
            columns_width = [max(len(val[col_idx]) for val in items) for col_idx, column in enumerate(headers)]
            columns_width = [max(columns_width[col_idx], len(column)) for col_idx, column in enumerate(headers)]

            print('|'. join('{name:{width}} '.format(name=column, width=columns_width[col_idx])
                            for col_idx, column in enumerate(headers)))
            for item in items:
                print('|'.join('{name:{width}} '.format(name=column, width=columns_width[col_idx])
                               for col_idx, column in enumerate(item)))


def undeploy_kubernetes(args):
    """
    Undeploy model to kubernetes

    :param args: command arguments with .model_id, .namespace
    :type args: :py:class:`argparse.Namespace`
    :return: None
    """
    edi_client = legion.external.edi.build_client(args)
    model_deployments = edi_client.inspect(args.model_id, args.model_version)

    if not model_deployments:
        if args.ignore_not_found:
            print('Cannot find any deployment - ignoring')
            return
        else:
            raise Exception('Cannot find any deployment')

    if len(model_deployments) > 1:
        raise Exception('Founded more then one deployment')

    target_deployment = model_deployments[0]

    edi_client.undeploy(args.model_id, args.grace_period, args.model_version)

    if not args.no_wait:
        start = time.time()

        while True:
            elapsed = time.time() - start
            if elapsed > args.wait_timeout and args.wait_timeout != 0:
                raise Exception('Time out: model has not been undeployed')

            information = [info
                           for info in edi_client.inspect()
                           if info.model == target_deployment.model and info.version == target_deployment.version]

            if not information:
                break

            time.sleep(1)


def scale_kubernetes(args):
    """
    Scale model instances

    :param args: command arguments with .model_id, .namespace and .scale
    :type args: :py:class:`argparse.Namespace`
    :return: None
    """
    edi_client = legion.external.edi.build_client(args)
    edi_client.scale(args.model_id, args.scale, args.model_version)


def deploy_kubernetes(args):
    """
    Deploy kubernetes model

    :param args: command arguments with .model_id, .namespace and .scale
    :type args: :py:class:`argparse.Namespace`
    :return: None
    """
    edi_client = legion.external.edi.build_client(args)
    edi_client.deploy(args.image, args.scale)

    # Start waiting of readiness of all PODs
    if not args.no_wait:
        start = time.time()

        while True:
            elapsed = time.time() - start
            if elapsed > args.wait_timeout and args.wait_timeout != 0:
                break

            information = [info for info in edi_client.inspect() if info.image == args.image]

            if not information:
                raise Exception('Cannot find model deployment after deploy for image {}'.format(args.image))

            deployment = information[0]

            if deployment.ready_replicas >= deployment.scale and deployment.model_api_ok:
                break

            time.sleep(1)



#
#    Copyright 2017 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
from setuptools import setup
import os
import re

PACKAGE_ROOT_PATH = os.path.dirname(os.path.abspath(__file__))


def extract_requirements(filename):
    """
    Extracts requirements from a pip formatted requirements file.
    :param filename: str path to file
    :return: list of package names as strings
    """
    with open(filename, 'r') as requirements_file:
        return requirements_file.read().splitlines()


def extract_version(filename):
    """
    Extract version from .py file using regex
    :param filename: str path to file
    :return: str version
    """
    with open(filename, 'r') as version_file:
        file_content = open(filename, "rt").read()
        VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
        mo = re.search(VSRE, file_content, re.M)
        if mo:
            return mo.group(1)
        else:
            raise RuntimeError("Unable to find version string in %s." % (file_content,))


setup(name='legion',
      version=extract_version(os.path.join(PACKAGE_ROOT_PATH, 'legion', 'version.py')),
      description='Legion',
      url='http://github.com/akharlamov/legion-root',
      author='Alexey Kharlamov, Kirill Makhonin',
      author_email='alexey@kharlamov.biz, kirill@makhonin.biz',
      license='Apache v2',
      packages=['legion'],
      include_package_data=True,
      scripts=['bin/legionctl', 'bin/edi', 'bin/legion-template'],
      install_requires=extract_requirements(os.path.join(PACKAGE_ROOT_PATH, 'requirements', 'base.txt')),
      test_suite='nose.collector',
      tests_require=extract_requirements(os.path.join(PACKAGE_ROOT_PATH, 'requirements', 'test.txt')),
      zip_safe=False)



#
#    Copyright 2017 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
Model metrics
"""
import os
import socket
import time
from enum import Enum

import legion.config
from legion.utils import normalize_name


class Metric(Enum):
    """
    Metric type
    """

    TRAINING_ACCURACY = 'training-accuracy'
    TEST_ACCURACY = 'test-accuracy'
    TRAINING_LOSS = 'training-loss'


def get_metric_endpoint():
    """
    Get metric endpoint

    :return: metric server endpoint
    """
    host = os.getenv(*legion.config.GRAPHITE_HOST)
    port = int(os.getenv(*legion.config.GRAPHITE_PORT))
    namespace = os.getenv(*legion.config.GRAPHITE_NAMESPACE)
    return host, port, namespace


def get_build_number():
    """
    Get current build number

    :return: int -- build number
    """
    try:
        return int(os.getenv(*legion.config.BUILD_NUMBER))
    except ValueError:
        raise Exception('Cannot parse build number as integer')


def get_metric_name(metric, model_id):
    """
    Get metric name on stats server

    :param metric: instance of Metric or custom name
    :type metric: :py:class:`legion.metrics.Metric` or str
    :param model_id: model ID
    :type model_id: str
    :return: str -- metric name on stats server
    """
    name = metric.value if isinstance(metric, Metric) else str(metric)
    return normalize_name('{}.metrics.{}'.format(model_id, name))


def get_build_metric_name(model_id):
    """
    Get build # name on stats server

    :param model_id: model ID
    :type model_id: str
    :return: str -- build # name on stats server
    """
    return '{}.metrics.build'.format(model_id)


def send_udp(host, port, message):
    """
    Send message with UDP

    :param host: target host
    :type host: str
    :param port: target port
    :type port: int
    :param message: data string or bytes
    :type message: str or bytes
    :return: None
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

    if isinstance(message, str):
        message = message.encode('utf-8')

    sock.sendto(message, (host, port))


def send_tcp(host, port, message):
    """
    Send message with TCP

    :param host: target host
    :type host: str
    :param port: target port
    :type port: int
    :param message: data string or bytes
    :type message: str or bytes
    :return: None
    """
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect((host, port))

    if isinstance(message, str):
        message = message.encode('utf-8')

    sock.send(message)
    sock.close()


def send_metric(model_id, metric, value):
    """
    Send build metric value

    :param model_id: model ID
    :type model_id: str
    :param metric: metric type or metric name
    :type metric: :py:class:`legion.metrics.Metric` or str
    :param value: metric value
    :type value: float or int
    :return: None
    """
    host, port, namespace = get_metric_endpoint()

    metric_name = '%s.%s' % (namespace, get_metric_name(metric, model_id))
    message = "%s %f %d\n" % (metric_name, float(value), int(time.time()))
    send_tcp(host, port, message)

    build_no = get_build_number()
    metric_name = '%s.%s' % (namespace, get_metric_name('build', model_id))
    message = "%s %f %d\n" % (metric_name, build_no, int(time.time()))
    send_tcp(host, port, message)



#
#    Copyright 2018 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
legion k8s properties class
"""
import base64
import logging
import time

import kubernetes
import kubernetes.client
import kubernetes.client.models
import kubernetes.client.rest
import kubernetes.config
import kubernetes.config.config_exception

import legion.containers.headers
import legion.config
import legion.model
import legion.k8s.watch
import legion.k8s.utils
import legion.k8s.services
import legion.utils

LOGGER = logging.getLogger(__name__)


class K8SPropertyStorage:
    """
    K8S property storage
    """

    def __init__(self, storage_name, k8s_namespace=None, data=None, cache_ttl=None):
        """
        Build K8S properties storage

        :param storage_name: name of storage
        :type storage_name: str
        :param k8s_namespace: k8s namespace of None for auto deducing (auto deducing works only in-cluster)
        :type k8s_namespace: str or None
        :param data: start data or None
        :type data: dict[str, any]
        :param cache_ttl: amount of lifetime of cache in seconds or None for disabling
        :type cache_ttl: int
        """
        if not data:
            data = {}
        self._state = data

        self._cache_ttl = cache_ttl

        self._storage_name = legion.utils.normalize_name(storage_name, dns_1035=True)

        self._last_load_time = None
        self._saved = False

        self._k8s_namespace = k8s_namespace

        LOGGER.info('Initializing {!r}'.format(self))

    @classmethod
    def retrive(cls, storage_name, k8s_namespace=None):
        """
        Initialize storage class from K8S object and load

        :param cls: linked class
        :type cls: :py:class:`legion.k8s.properties.K8SPropertyStorage`
        :param storage_name: name of storage
        :type storage_name: str
        :param k8s_namespace: k8s namespace of None for auto deducing (auto deducing works only in-cluster)
        :type k8s_namespace: str or None
        :return: :py:class:`legion.k8s.properties.K8SPropertyStorage` -- built and initialized object
        """
        instance = cls(storage_name, k8s_namespace)
        instance.load()
        return instance

    @classmethod
    def list(cls, k8s_namespace=None):
        """
        Get list of storage names

        :param cls: linked class
        :type cls: :py:class:`legion.k8s.properties.K8SPropertyStorage`
        :param k8s_namespace: target namespace or None for auto deducing (auto deducing works only in-cluster)
        :type k8s_namespace: str
        :return: list[str] -- list of storages
        """
        if not k8s_namespace:
            k8s_namespace = legion.k8s.get_current_namespace()

        resources = cls._find_k8s_resources(k8s_namespace)
        names = [
            resource.metadata.labels[legion.containers.headers.DOMAIN_MODEL_PROPERTY_TYPE]
            for resource in resources
        ]
        return names

    @property
    def k8s_name(self):
        """
        Get storage name (object name)

        :return: str -- storage name
        """
        return self._storage_name

    @property
    def k8s_namespace(self):
        """
        Get K8S namespace name

        :return: str -- K8S namespace name
        """
        return self._k8s_namespace

    @property
    def k8s_namespace_or_default(self):
        """
        Get K8S namespace or set and retrive default namespace

        :return: str -- setted or default namespace
        """
        if not self._k8s_namespace:
            LOGGER.info('Detecting default namespace')
            self._k8s_namespace = legion.k8s.utils.get_current_namespace()

        return self._k8s_namespace

    @property
    def data(self):
        """
        Get current state

        :return: dict[str, str] -- current state
        """
        self._check_and_reload()
        return self._state

    @data.setter
    def data(self, new_state):
        """
        Update current state with another dict

        :param new_state: new state
        :type new_state: dict[str, str-serializable]
        :return: None
        """
        self._last_load_time = None
        self._saved = False
        self._state = {k: str(v) for (k, v) in new_state.items()}

    @data.deleter
    def data(self):
        """
        Reset current state

        :return: None
        """
        self._last_load_time = None
        self._saved = False
        self._state = {}

    def __setitem__(self, key, value):
        """
        Set property value (it would be converted to string immediately)

        :param key: key (property name)
        :type key: str
        :param value: value of property, should be str serializable
        :type value: str serializable object
        :return: None
        """
        self._last_load_time = None
        self._saved = False
        self._state[key] = str(value)

    def __getitem__(self, key):
        """
        Get value of property

        :param key: name of property field
        :type key: str
        :return: any -- property value or None
        """
        self._check_and_reload()
        return self._state.get(key)

    def keys(self):
        """
        Get list of keys

        :return: list[str] -- list of keys
        """
        self._check_and_reload()
        return list(self._state.keys())

    def get(self, key, cast):
        """
        Get value of property with cast to legion type

        :param key: name of property field
        :type key: str
        :param cast:
        :type cast: LegionType
        :return:
        """
        value = self[key]
        legion_base_type = cast.representation_type
        return legion_base_type.parse(value)

    @property
    def _core_api(self):
        """
        Build K8S Core API instance

        :return: :py:class:`kubernetes.client.CoreV1Api` -- Core V1 API instance
        """
        client = legion.k8s.utils.build_client()
        return kubernetes.client.CoreV1Api(client)

    def _read_data_from_dict(self, source_dict):
        """
        Update current state with values from dict

        :param source_dict: dict with readed values
        :type source_dict: dict[str, str]
        :return: None
        """
        self._state = source_dict

    def _write_data_to_dict(self):
        """
        Get dict of current state to be saved in K8S

        :return: dict[str, str] -- dict with values to save
        """
        return self._state

    @staticmethod
    def _find_k8s_resources(k8s_namespace):
        """
        Virtual function. Find K8S

        :return: V1ConfigMap or V1Secret  -- K8S object
        """
        raise NotImplementedError('This function is not implemented in this class')

    def _read_k8s_resource(self):
        """
        Virtual function. Get appropriate K8S object

        :return: V1ConfigMap or V1Secret  -- K8S object
        """
        raise NotImplementedError('This function is not implemented in this class')

    def _read_k8s_resource_exception_handler(self, exception):
        """
        Virtual function. Handler for exceptions that can appear during reading K8S resource

        :param exception: appeared exception
        :type exception: :py:class:`Exception`
        :return: None
        """
        raise NotImplementedError('This function is not implemented in this class')

    def load(self):
        """
        Load data from K8S

        :return: None
        """
        try:
            LOGGER.debug('Reading {} {!r} in namespace {!r}'.format(self.__class__.__name__,
                                                                    self.k8s_name,
                                                                    self.k8s_namespace_or_default))

            config_map_object = self._read_k8s_resource()

            self._read_data_from_dict(config_map_object.data)
        except Exception as load_exception:
            self._read_k8s_resource_exception_handler(load_exception)

        self._last_load_time = time.time()
        self._saved = True

    def _check_and_reload(self):
        """
        Check and reload data if last operation was reading and time is out

        :return: None
        """
        if not self._cache_ttl:
            return

        if not self._last_load_time or not self._saved:
            return

        delta = time.time() - self._last_load_time
        if delta > self._cache_ttl:
            self.load()

    def _write_k8s_resource(self):
        """
        Virtual function. Create K8S object with appropriate content (from self._state)

        :return: None
        """
        raise NotImplementedError('This function is not implemented in this class')

    def _write_k8s_resource_exception_handler(self, exception):
        """
        Virtual function. Handler for exceptions that can appear during creating K8S resource

        :param exception: appeared exception
        :type exception: :py:class:`Exception`
        :return: None
        """
        raise NotImplementedError('This function is not implemented in this class')

    def save(self):
        """
        Write current state object (self._state) to K8S object

        :return: None
        """
        try:
            LOGGER.debug('Writing {!r}'.format(self))
            self._write_k8s_resource()
        except Exception as write_exception:
            self._write_k8s_resource_exception_handler(write_exception)

        self._saved = True

    def _remove_k8s_resource(self, delete_options):
        """
        Remove appropriate resource from K8S

        :param delete_options: K8S delete options
        :type delete_options: :py:class:`kubernetes.client.V1DeleteOptions`
        :return: None
        """
        raise NotImplementedError('This function is not implemented in this class')

    def destroy(self):
        """
        Remove appropriate K8S object

        :return: None
        """
        delete_options = kubernetes.client.V1DeleteOptions(propagation_policy='Foreground',
                                                           grace_period_seconds=0)

        LOGGER.debug('Deleting {!r}'.format(self))
        self._remove_k8s_resource(delete_options)

    def __repr__(self):
        """
        Get representation of object

        :return: str -- representation
        """
        return '<{} k8s_name={!r}, k8s_namespace={!r}, data={!r}>' \
            .format(self.__class__.__name__, self._storage_name, self._k8s_namespace, self._state)


class K8SConfigMapStorage(K8SPropertyStorage):
    """
    Storage for properties that uses K8S config map for storing
    """

    @staticmethod
    def _find_k8s_resources(k8s_namespace):
        """
        Find K8S resources

        :return: :py:class:`kubernetes.client.V1ConfigMap` -- K8S object
        """
        client = legion.k8s.utils.build_client()
        core_api = kubernetes.client.CoreV1Api(client)
        objects = core_api.list_namespaced_config_map(k8s_namespace).items

        objects = [
            obj
            for obj in objects
            if obj.metadata.labels and legion.containers.headers.DOMAIN_MODEL_PROPERTY_TYPE in obj.metadata.labels
        ]

        return objects

    def _read_k8s_resource(self):
        """
        Get appropriate K8S config map

        :return: :py:class:`kubernetes.client.V1ConfigMap` -- K8S object
        """
        return self._core_api.read_namespaced_config_map(self.k8s_name, self.k8s_namespace_or_default)

    def _read_k8s_resource_exception_handler(self, exception):
        """
        Handle exceptions that can appear during reading K8S config map

        :param exception: appeared exception
        :type exception: :py:class:`Exception`
        :return: None
        """
        raise Exception('Cannot read config map {!r} in namespace {!r}: {}'
                        .format(self.k8s_name, self.k8s_namespace_or_default, exception))

    def _write_k8s_resource(self):
        """
        Create K8S config map with appropriate content (from self._state)

        :return: None
        """
        config_map_metadata = kubernetes.client.models.v1_object_meta.V1ObjectMeta(
            name=self.k8s_name,
            labels={
                legion.containers.headers.DOMAIN_MODEL_PROPERTY_TYPE: self._storage_name
            }
        )

        config_map_body = kubernetes.client.models.v1_config_map.V1ConfigMap(
            metadata=config_map_metadata,
            data=self._write_data_to_dict()
        )

        try:
            LOGGER.info('Overriding ConfigMap {!r} in {!r} namespace'.format(self.k8s_name,
                                                                             self.k8s_namespace_or_default))
            self._core_api.replace_namespaced_config_map(self.k8s_name,
                                                         self.k8s_namespace_or_default,
                                                         config_map_body)
        except kubernetes.client.rest.ApiException as invoke_exception:
            if invoke_exception.status == 404:
                LOGGER.info('Creating ConfigMap {!r} in {!r} namespace'.format(self.k8s_name,
                                                                               self.k8s_namespace_or_default))
                self._core_api.create_namespaced_config_map(self.k8s_namespace_or_default,
                                                            config_map_body)

    def _write_k8s_resource_exception_handler(self, exception):
        """
        Handle exceptions that can appear during creating K8S resource

        :param exception: appeared exception
        :type exception: :py:class:`Exception`
        :return: None
        """
        raise Exception('Cannot write config map {!r} in namespace {!r}: {}'
                        .format(self.k8s_name, self.k8s_namespace_or_default, exception))

    def _remove_k8s_resource(self, delete_options):
        """
        Remove config map from K8S

        :param delete_options: K8S delete options
        :type delete_options: :py:class:`kubernetes.client.V1DeleteOptions`
        :return: None
        """
        self._core_api.delete_namespaced_config_map(self.k8s_name,
                                                    self.k8s_namespace_or_default,
                                                    body=delete_options)


class K8SSecretStorage(K8SPropertyStorage):
    """
    Storage for properties that uses K8S secret for storing
    """

    def _read_data_from_dict(self, source_dict):
        """
        Update current state with values from dict

        :param source_dict: dict with readed values
        :type source_dict: dict[str, str]
        :return: None
        """
        self._state = {
            k: base64.b64decode(v.encode('ascii')).decode('utf-8')
            for k, v in source_dict.items()
        }

    def _write_data_to_dict(self):
        """
        Get dict of current state to be saved in K8S

        :return: dict[str, str] -- dict with values to save
        """
        return {
            k: base64.b64encode(v.encode('utf-8')).decode('ascii')  # encode string with base64
            for k, v in self._state.items()
        }

    @staticmethod
    def _find_k8s_resources(k8s_namespace):
        """
        Find K8S resources

        :return: :py:class:`kubernetes.client.V1Secret` -- K8S object
        """
        client = legion.k8s.utils.build_client()
        core_api = kubernetes.client.CoreV1Api(client)
        objects = core_api.list_namespaced_secret(k8s_namespace).items

        objects = [
            obj
            for obj in objects
            if obj.metadata.labels and legion.containers.headers.DOMAIN_MODEL_PROPERTY_TYPE in obj.metadata.labels
        ]

        return objects

    def _read_k8s_resource(self):
        """
        Get appropriate K8S secret

        :return: :py:class:`kubernetes.client.V1Secret` -- K8S object
        """
        return self._core_api.read_namespaced_secret(self.k8s_name, self.k8s_namespace_or_default)

    def _read_k8s_resource_exception_handler(self, exception):
        """
        Handle exceptions that can appear during reading K8S secret

        :param exception: appeared exception
        :type exception: :py:class:`Exception`
        :return: None
        """
        raise Exception('Cannot read secret {!r} in namespace {!r}: {}'
                        .format(self.k8s_name, self.k8s_namespace_or_default, exception))

    def _write_k8s_resource(self):
        """
        Create K8S secret with appropriate content (from self._state)

        :return: None
        """
        secret_metadata = kubernetes.client.models.v1_object_meta.V1ObjectMeta(
            name=self.k8s_name,
            labels={
                legion.containers.headers.DOMAIN_MODEL_PROPERTY_TYPE: self._storage_name
            }
        )

        secret_body = kubernetes.client.models.v1_secret.V1Secret(
            metadata=secret_metadata,
            data=self._write_data_to_dict()
        )

        try:
            LOGGER.info(
                'Overriding Secret {!r} in {!r} namespace'.format(self.k8s_name, self.k8s_namespace_or_default))
            self._core_api.replace_namespaced_secret(self.k8s_name, self.k8s_namespace_or_default, secret_body)
        except kubernetes.client.rest.ApiException as invoke_exception:
            if invoke_exception.status == 404:
                LOGGER.info(
                    'Creating Secret {!r} in {!r} namespace'.format(self.k8s_name, self.k8s_namespace_or_default))
                self._core_api.create_namespaced_secret(self.k8s_namespace_or_default, secret_body)

    def _write_k8s_resource_exception_handler(self, exception):
        """
        Handle exceptions that can appear during creating K8S resource

        :param exception: appeared exception
        :type exception: :py:class:`Exception`
        :return: None
        """
        raise Exception('Cannot write Secret {!r} in namespace {!r}: {}'
                        .format(self.k8s_name, self.k8s_namespace_or_default, exception))

    def _remove_k8s_resource(self, delete_options):
        """
        Remove secret from K8S

        :param delete_options: K8S delete options
        :type delete_options: :py:class:`kubernetes.client.V1DeleteOptions`
        :return: None
        """
        self._core_api.delete_namespaced_secret(self.k8s_name,
                                                self.k8s_namespace_or_default,
                                                body=delete_options)



#
#    Copyright 2018 EPAM Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#
"""
legion k8s enclave class
"""
import logging
import os
import json

import kubernetes
import kubernetes.client
import kubernetes.config
import kubernetes.config.config_exception
import urllib3
import urllib3.exceptions

import legion.containers.headers
import legion.config
from legion.k8s.definitions import ENCLAVE_NAMESPACE_LABEL
from legion.k8s.definitions import \
    LEGION_COMPONENT_NAME_API, LEGION_COMPONENT_NAME_EDI, \
    LEGION_COMPONENT_NAME_GRAFANA, LEGION_COMPONENT_NAME_GRAPHITE
import legion.k8s.watch
import legion.k8s.utils
import legion.k8s.services
import legion.k8s.properties
import legion.utils

LOGGER = logging.getLogger(__name__)


class Enclave:
    """l
    Contains overall information about enclave, its services and models
    """

    def __init__(self, name):
        """
        Create an Enclave instance

        :param name: enclave name
        """
        self._name = name

        self._data_loaded = False

        self._edi_service = None
        self._api_service = None
        self._grafana_service = None
        self._graphite_service = None

    @staticmethod
    def build_from_namespace_object(ns_object):
        """
        Build Enclave from namespace object

        :param ns_object: namespace object
        :return: :py:class:`Enclave` -- enclave
        """
        return Enclave(ns_object.metadata.name)

    @staticmethod
    def is_enclave(namespace):
        """
        Is specific K8S namespace is K8S namespace

        :param namespace: K8S namespace
        :return: bool - is namespace is enclave namespace
        """
        return namespace.metadata.labels and ENCLAVE_NAMESPACE_LABEL in namespace.metadata.labels

    def __repr__(self):
        """
        Return representation of object

        :return: str -- string representation
        """
        return 'Enclave({!r})'.format(self._name)

    def _load_services(self):
        """
        Load all services data (lazy loading)

        :return: None
        """
        if self._data_loaded:
            return

        self._data_loaded = True

        self._edi_service = legion.k8s.services.get_service(self.name, LEGION_COMPONENT_NAME_EDI)
        self._api_service = legion.k8s.services.get_service(self.name, LEGION_COMPONENT_NAME_API)
        self._grafana_service = legion.k8s.services.get_service(self.name, LEGION_COMPONENT_NAME_GRAFANA)
        self._graphite_service = legion.k8s.services.get_service(self.name, LEGION_COMPONENT_NAME_GRAPHITE)

    @property
    def name(self):
        """
        Return enclave name, currently it equals to K8s namespace name

        :return: str -- enclave name
        """
        return self._name

    @property
    def namespace(self):
        """
        Return enclave namespace, currently it equals to enclave name

        :return: str -- enclave namespace
        """
        return self._name

    @property
    def edi_service(self):
        """
        Return EDI service object.

        :return: :py:class:`legion.k8s.Service` -- EDI service object or None
        """
        self._load_services()
        return self._edi_service

    @property
    def api_service(self):
        """
        Return model API service object (previously known as EDGE).

        :return: :py:class:`legion.k8s.Service` -- api service object or None
        """
        self._load_services()
        return self._api_service

    @property
    def grafana_service(self):
        """
        Return Grafana service object.

        :return: :py:class:`legion.k8s.Service` -- Grafana service object or None
        """
        self._load_services()
        return self._grafana_service

    @property
    def graphite_service(self):
        """
        Return Graphite service object.

        :return: :py:class:`legion.k8s.Service` -- Graphite service object or None
        """
        self._load_services()
        return self._graphite_service

    @property
    def models(self):
        """
        Return a dict of Models services.

        :return: dict[:py:class:`legion.k8s.ModelIdVersion`, :py:class:`legion.k8s.ModelService`] -- deployed models
        """
        models = {}
        for service in legion.k8s.services.find_all_services(namespace=self.name):
            if legion.k8s.services.ModelService.is_model_service(service):
                model_service = legion.k8s.services.ModelService(service)
                models[model_service.id_and_version] = model_service
        return models

    def get_models(self, model_id, model_version=None):
        """
        Get models that fit match criterions (model id and model version, model id may be *, model version may be *)

        :param model_id: model id or */None for all models
        :type model_id: str
        :param model_version: (Optional) model version or */None for all
        :type model_version: str
        :return: list[:py:class:`legion.k8s.ModelService`] -- founded model services
        """
        return [
            model_service
            for id_and_version, model_service in self.models.items()
            if model_id in (id_and_version.id, '*', None) and model_version in (id_and_version.version, None, '*')
        ]

    def get_models_strict(self, model_id, model_version=None):
        """
        Get models that fit match criterions (model id and model version, model id may be *, model version may be *)
        If more that one model would be found with unstrict criterion - exception would be raised
        If no one model would be found - exception would be raised

        param model_id: model id or * for all models
        :type model_id: str
        :param model_version: (Optional) model version
        :type model_version: str
        :return: list[:py:class:`legion.k8s.ModelService`] -- founded model services
        """
        model_services = self.get_models(model_id, model_version)

        if len(model_services) > 1 and not model_version:
            raise Exception('Please specify version of model')

        if not model_services:
            raise Exception('No one model can be found')

        return model_services

    def _validate_model_properties_storage(self, model_id, model_version, properties, default_values):
        """
        Validate that model properties for model exists in a cluster and contains required properties
        If there are not properties storage in a cluster - create with default values

        :param model_id: model ID
        :type model_id: str
        :param model_version: model version
        :type model_version: str
        :param properties: required properties or None
        :type properties: list[str] or None
        :param default_values: default values for properties
        :type default_values: dict[str, str]
        :return: None
        """
        if not properties:  # if model does not require properties check can be omitted
            return

        registered_storages = legion.k8s.properties.K8SConfigMapStorage.list(self.namespace)
        storage_name = legion.utils.model_properties_storage_name(model_id, model_version)
        storage = legion.k8s.K8SConfigMapStorage(storage_name, self.namespace)

        if storage_name in registered_storages:
            LOGGER.info('Analyzing properties storage {!r} for model {!r}'.format(storage, model_id))
            storage.load()

            missed_properties = set(properties) - set(storage.keys())

            for missed_property in missed_properties:
                if missed_property not in default_values:
                    raise Exception('Cannot find default value for property {}'.format(missed_property))

                storage[missed_property] = default_values[missed_property]
                LOGGER.info('Property {!r} has been set to default value {!r}'.format(missed_property,
                                                                                      default_values[missed_property]))
        else:
            LOGGER.info('Creating properties storage {!r} for model {!r} with default values'.format(storage, model_id))
            for k, v in default_values.items():
                storage[k] = v
                LOGGER.info('Property {!r} has been set to default value {!r}'.format(k, v))
                storage.save()

    def deploy_model(self, image, count=1):
        """
        Deploy new model

        :param image: docker image with model
        :type image: str
        :param count: count of pods
        :type count: int
        :return: :py:class:`legion.k8s.services.ModelService` -- model service
        """
        if count < 1:
            raise Exception('Invalid scale parameter: should be greater then 0')

        client = legion.k8s.utils.build_client()
        labels = legion.k8s.utils.get_docker_image_labels(image)
        k8s_name, compatible_labels, model_id, model_version = legion.k8s.utils.get_meta_from_docker_labels(labels)

        model_properties = labels.get(legion.containers.headers.DOMAIN_MODEL_PROPERTIES)
        model_properties_default_values = labels.get(legion.containers.headers.DOMAIN_MODEL_PROPERTY_VALUES)

        self._validate_model_properties_storage(
            model_id,
            model_version,
            model_properties.split(',') if model_properties else [],
            json.loads(model_properties_default_values) if model_properties_default_values else {}
        )

        if self.get_models(model_id, model_version):
            raise Exception('Duplicating model id and version (id={}, version={})'.format(model_id, model_version))

        # REFACTOR, maybe we should remove that
        container_env_variables = {
            legion.config.STATSD_HOST[0]: self.graphite_service.internal_domain,
            legion.config.STATSD_PORT[0]: str(self.graphite_service.internal_port)
        }

        container = kubernetes.client.V1Container(
            name='model',
            image=image,
            image_pull_policy='Always',
            env=[
                kubernetes.client.V1EnvVar(name=k, value=str(v))
                for k, v in container_env_variables.items()
            ],
            ports=[kubernetes.client.V1ContainerPort(container_port=5000, name='api', protocol='TCP')])

        pod_template = kubernetes.client.V1PodTemplateSpec(
            metadata=kubernetes.client.V1ObjectMeta(labels=compatible_labels),
            spec=kubernetes.client.V1PodSpec(
                containers=[container],
                service_account_name=legion.config.MODEL_INSTANCE_SERVICE_ACCOUNT_NAME
            ))

        deployment_spec = kubernetes.client.ExtensionsV1beta1DeploymentSpec(
            replicas=count,
            template=pod_template)

        deployment = kubernetes.client.ExtensionsV1beta1Deployment(
            api_version="extensions/v1beta1",
            kind="Deployment",
            metadata=kubernetes.client.V1ObjectMeta(name=k8s_name, labels=compatible_labels),
            spec=deployment_spec)

        extensions_v1beta1 = kubernetes.client.ExtensionsV1beta1Api(client)

        LOGGER.info('Creating deployment {} in namespace {}'.format(k8s_name, self.namespace))
        extensions_v1beta1.create_namespaced_deployment(
            body=deployment,
            namespace=self.namespace)

        # Creating a service
        service_selector = {k: v for k, v in compatible_labels.items()
                            if k in [legion.containers.headers.DOMAIN_MODEL_ID,
                                     legion.containers.headers.DOMAIN_MODEL_VERSION]}

        service_spec = kubernetes.client.V1ServiceSpec(
            selector=service_selector,
            ports=[kubernetes.client.V1ServicePort(name='api', protocol='TCP', port=5000, target_port='api')])

        service = kubernetes.client.V1Service(
            api_version='v1',
            kind='Service',
            metadata=kubernetes.client.V1ObjectMeta(name=k8s_name, labels=compatible_labels),
            spec=service_spec)

        core_v1api = kubernetes.client.CoreV1Api(client)

        LOGGER.info('Creating service {} in namespace {}'.format(k8s_name, self.namespace))
        core_v1api.create_namespaced_service(
            body=service,
            namespace=self.namespace)

        model_services = self.get_models(model_id, model_version)
        if model_services:
            return model_services[0]
        else:
            raise Exception('Cannot find created model service for model {} id {}'.format(model_id, model_version))

    def watch_models(self):
        """
        Watch for models update events in this Enclave object

        :return: (str, :py:class:`legion.k8s.services.ModelService`) -- (event type [ADDED, DELETED, MODIFIED], model)
        """
        for event_type, service in self.watch_services():
            if legion.k8s.services.ModelService.is_model_service(service.k8s_service):
                model_service = legion.k8s.services.ModelService(service.k8s_service)
                yield (event_type, model_service)

    def watch_model_service_endpoints_state(self):
        """
        Watch for model service endpoints state

        :return: (list[:py:class:`legion.k8s.services.ModelServiceEndpoint`] -- new version of model service endpoints
        """
        endpoints = set()

        def normalize_endpoints(endpoints_to_normalize):
            """
            Normalize endpoints set to list with default endpoints for each models

            :param endpoints_to_normalize: source model endpoints
            :type endpoints_to_normalize: set[:py:class:`legion.k8s.services.ModelServiceEndpoint`]
            :return list[:py:class:`legion.k8s.services.ModelServiceEndpoint`] -- endpoints with default
            """
            result = []
            default_endpoints_models = {}

            for endpoint in endpoints_to_normalize:
                model_id = endpoint.model_service.id

                # TODO: Change in future, add default model choosing algorithm
                default_endpoints_models[model_id] = endpoint
                result.append(endpoint)

            # Append defaults to the end
            for default_endpoint in default_endpoints_models.values():
                result.append(default_endpoint.build_default())

            return result

        for (event_type, model_service) in self.watch_models():
            model_endpoint = legion.k8s.services.ModelServiceEndpoint(model_service)

            if event_type == legion.k8s.EVENT_ADDED:
                endpoints.add(model_endpoint)
            elif event_type == legion.k8s.EVENT_MODIFIED:
                endpoints.add(model_endpoint)
            elif event_type == legion.k8s.EVENT_DELETED:
                if model_endpoint in endpoints:
                    endpoints.remove(model_endpoint)
            else:
                LOGGER.error('Got unknown event type: {}'.format(event_type))

            yield normalize_endpoints(endpoints)

    def watch_services(self):
        """
        Generate which returns events for services updates

        :return: (str, :py:class:`legion.k8s.services.Service`) -- (event type [ADDED, DELETED, MODIFIED], service)
        """
        client = legion.k8s.utils.build_client()

        core_api = kubernetes.client.CoreV1Api(client)

        with legion.k8s.watch.ResourceWatch(core_api.list_namespaced_service,
                                            namespace=self.namespace,
                                            filter_callable=legion.k8s.services.Service.is_legion_service,
                                            object_constructor=legion.k8s.services.Service) as watch:
            for (event_type, event_object) in watch.stream:
                yield (event_type, event_object)

    @staticmethod
    def watch_enclaves():
        """
        Generate which returns events for enclaves

        :return: A tuple (event type [ADDED, DELETED, MODIFIED], service object)
        :rtype: (str, Enclave)
        """
        client = legion.k8s.utils.build_client()

        core_api = kubernetes.client.CoreV1Api(client)

        with legion.k8s.watch.ResourceWatch(core_api.list_namespace,
                                            filter_callable=Enclave.is_enclave,
                                            object_constructor=Enclave.build_from_namespace_object) as watch:
            for (event_type, event_object) in watch.stream:
                yield (event_type, event_object)

    def delete(self, grace_period_seconds=0):
        """
        Delete enclave

        :param grace_period_seconds: grace period in seconds
        :type grace_period_seconds: int
        :return: None
        """
        client = legion.k8s.utils.build_client()
        core_api = kubernetes.client.CoreV1Api(client)

        body = kubernetes.client.V1DeleteOptions(propagation_policy='Foreground',
                                                 grace_period_seconds=grace_period_seconds)

        core_api.delete_namespace(self.name, body)


def find_enclaves():
    """
    Get a list of enclaves

    :return list[Enclave] -- list of found enclaves
    """
    client = legion.k8s.utils.build_client()

    core_api = kubernetes.client.CoreV1Api(client)
    all_namespaces = core_api.list_namespace()

    enclaves = []
    for namespace in all_namespaces.items:
        if namespace.metadata.labels and ENCLAVE_NAMESPACE_LABEL in namespace.metadata.labels:
            enclaves.append(legion.k8s.enclave.Enclave(namespace.metadata.name))

    return enclaves
